use std::{
    future::Future,
    pin::Pin,
    sync::{
        atomic::{AtomicBool, Ordering::Relaxed},
        Arc,
    },
    task::{Context, Poll, Waker},
};

use crate::{utils::try_ready, volmutex::VolMutex};

/// Notification receiver. Await it to wait for notification.
pub struct Slave(Arc<State>);

struct State {
    should_wake_up: AtomicBool,
    waker: VolMutex<Option<Waker>>,
}

impl Future for Slave {
    type Output = ();

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        let state = &*self.get_mut().0;
        if state.should_wake_up.swap(false, Relaxed) {
            return Poll::Ready(());
        }

        {
            let Some(mut locked) = state.waker.try_lock() else {
                state.should_wake_up.store(false, Relaxed);
                return Poll::Ready(());
            };

            let maybe_waker = &mut *locked;
            if let Some(waker) = maybe_waker {
                waker.clone_from(cx.waker());
            } else {
                *maybe_waker = Some(cx.waker().clone());
            }
        }

        try_ready(&state.should_wake_up)
    }
}

pub struct Master {
    slave: Arc<State>,
    stored_waker: Option<Waker>,
}

impl Master {
    /// Notify waiter. If [`Slave`] wasn't listening to the notification right now,
    /// next await would return immediately. In tokio terms, wake-up permit is stored,
    /// only one wake-up permit is kept.
    pub fn notify(&mut self) {
        self.slave.should_wake_up.store(true, Relaxed);
        {
            let Some(locked) = self.slave.waker.try_lock() else {
                return;
            };
            self.stored_waker.clone_from(&*locked);
        }

        if let Some(waker) = &self.stored_waker {
            waker.wake_by_ref();
        }
    }
}

/// Make master/slave pair.
pub fn make() -> (Master, Slave) {
    let state = Arc::new(State {
        should_wake_up: AtomicBool::new(false),
        waker: VolMutex::new(None),
    });
    let master = Master {
        slave: Arc::clone(&state),
        stored_waker: None,
    };
    let slave = Slave(state);

    (master, slave)
}
